#![allow(clippy::undocumented_unsafe_blocks)]

use std::{mem, ptr};

use nix::{
    self,
    errno::Errno,
    libc::{self, c_ulong, signal, sigset_t, SIG_DFL, SIG_SETMASK},
    unistd::close,
};
use sendfd::SendWithFd;

use crate::unshare::{caps, error::ErrorCode as Err, run::ChildInfo};

unsafe fn fail(code: Err) -> ! {
    fail_errno(code, nix::errno::errno())
}
unsafe fn fail_errno(code: Err, errno: i32) -> ! {
    let errnox = Errno::from_i32(errno);
    eprintln!("exec error: {code:?} errno:{errnox}");
    libc::_exit(errno);
}

macro_rules! fail_safe {
    ($child:expr, $error:expr) => {
        drop($child);
        unsafe { fail($error) };
    };
}

macro_rules! fail_errno_safe {
    ($child:expr, $error:expr, $errno:expr) => {
        drop($child);
        unsafe { fail_errno($error, $errno) };
    };
}

pub const PR_SET_PDEATHSIG: nix::libc::c_int = 1;

#[allow(clippy::cognitive_complexity)]
pub fn child_after_clone(child: ChildInfo) -> ! {
    if let Some(&sig) = child.cfg.death_sig.as_ref() {
        if unsafe { libc::prctl(PR_SET_PDEATHSIG, sig as c_ulong, 0, 0, 0) } != 0 {
            fail_safe!(child, Err::ParentDeathSignal);
        }
    }

    //if let Some(ref drop_caps) = child.cfg.drop_caps {
    for cap in CAPS_DROP {
        if unsafe { libc::prctl(libc::PR_CAPBSET_DROP, cap, 0, 0, 0) } != 0 {
            match Errno::last() {
                Errno::EINVAL | Errno::EPERM => {
                    // We either don't have CAP_SETPCAP capability,
                    // or capabilities are not enabled in the kernel.
                    // Let's move on.
                }
                _ => {
                    // Genuine error condition, let's fail.
                    fail_safe!(child, Err::DropCap);
                }
            };
        }
    }

    if child.cfg.restore_sigmask {
        unsafe {
            let mut sigmask: sigset_t = mem::zeroed();
            libc::sigemptyset(&mut sigmask);
            libc::pthread_sigmask(SIG_SETMASK, &sigmask, ptr::null_mut());
            for sig in 1..32 {
                signal(sig, SIG_DFL);
            }
        }
    }

    if let Some(callback) = &child.pre_exec {
        if let Err(e) = callback() {
            fail_errno_safe!(child, Err::PreExec, e as i32);
        }
    }

    if let Some(seccomp_filter) = &child.seccomp_filter {
        if seccomp_filter.load().is_err() {
            fail_safe!(child, Err::Seccomp);
        } else if let Some(seccomp_sender) = &child.seccomp_sender {
            let fd = match seccomp_filter.get_notify_fd() {
                Ok(fd) => fd,
                Err(_) => {
                    fail_safe!(child, Err::Seccomp);
                }
            };
            if seccomp_sender.send_with_fd(&[0; 1], &[fd]).is_err() {
                fail_safe!(child, Err::SeccompNotify);
            }
            let _ = close(fd);
        }
    }

    unsafe { libc::execvp(child.filename, child.args.as_ptr()) };
    fail_safe!(child, Err::Exec);
}

// The list of Linux capabilities(7) to drop on startup.
// TODO: Make this configurable.
const CAPS_DROP: [caps::Capability; 36] = [
    caps::Capability::CAP_AUDIT_READ,
    caps::Capability::CAP_AUDIT_WRITE,
    caps::Capability::CAP_AUDIT_CONTROL,
    caps::Capability::CAP_BLOCK_SUSPEND,
    caps::Capability::CAP_BPF,
    caps::Capability::CAP_DAC_OVERRIDE,
    caps::Capability::CAP_DAC_READ_SEARCH,
    caps::Capability::CAP_FOWNER,
    caps::Capability::CAP_FSETID,
    caps::Capability::CAP_SETGID,
    caps::Capability::CAP_SETUID,
    caps::Capability::CAP_LINUX_IMMUTABLE,
    caps::Capability::CAP_IPC_LOCK,
    caps::Capability::CAP_SYS_CHROOT,
    caps::Capability::CAP_SYS_PTRACE,
    caps::Capability::CAP_SYS_PACCT,
    caps::Capability::CAP_MAC_ADMIN,
    caps::Capability::CAP_MAC_OVERRIDE,
    caps::Capability::CAP_MKNOD,
    caps::Capability::CAP_LEASE,
    caps::Capability::CAP_NET_ADMIN,
    caps::Capability::CAP_NET_RAW,
    caps::Capability::CAP_PERFMON,
    caps::Capability::CAP_SYSLOG,
    caps::Capability::CAP_SYS_ADMIN,
    caps::Capability::CAP_SYS_BOOT,
    caps::Capability::CAP_SYS_CHROOT,
    caps::Capability::CAP_SYS_MODULE,
    caps::Capability::CAP_SYS_NICE,
    caps::Capability::CAP_SYS_PACCT,
    caps::Capability::CAP_SYS_PTRACE,
    caps::Capability::CAP_SYS_RAWIO,
    caps::Capability::CAP_SYS_RESOURCE,
    caps::Capability::CAP_SYS_TIME,
    caps::Capability::CAP_WAKE_ALARM,
    caps::Capability::CAP_SETPCAP,
];
